################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    string_buffer,
    BOOL,
    DWORD,
    HANDLE,
    HINSTANCE,
    HRESULT,
    HWND,
    INT,
    PINT,
    POINTER,
    PPWSTR,
    PVOID,
    PWSTR,
    )
from . import (
    INFINITE,
    SW_SHOW,
    MAX_PATH,
    SEE_MASK_NOCLOSEPROCESS,
    PROCESS_CREATE_PROCESS,
    PROC_THREAD_ATTRIBUTE_PARENT_PROCESS,
    CREATE_NEW_CONSOLE,
    EXTENDED_STARTUPINFO_PRESENT,
    cmdline_from_args,
    raise_on_zero,
    raise_on_hr,
    fun_fact,
    kernel,
    user,
    ref,
    )

_sh = ctypes.WinDLL("shell32.dll", use_last_error=True)

################################################################################

CSIDL_DESKTOP                 = 0x0000
CSIDL_INTERNET                = 0x0001
CSIDL_PROGRAMS                = 0x0002
CSIDL_CONTROLS                = 0x0003
CSIDL_PRINTERS                = 0x0004
CSIDL_PERSONAL                = 0x0005
CSIDL_FAVORITES               = 0x0006
CSIDL_STARTUP                 = 0x0007
CSIDL_RECENT                  = 0x0008
CSIDL_SENDTO                  = 0x0009
CSIDL_BITBUCKET               = 0x000a
CSIDL_STARTMENU               = 0x000b
CSIDL_MYDOCUMENTS             = CSIDL_PERSONAL
CSIDL_MYMUSIC                 = 0x000d
CSIDL_MYVIDEO                 = 0x000e
CSIDL_DESKTOPDIRECTORY        = 0x0010
CSIDL_DRIVES                  = 0x0011
CSIDL_NETWORK                 = 0x0012
CSIDL_NETHOOD                 = 0x0013
CSIDL_FONTS                   = 0x0014
CSIDL_TEMPLATES               = 0x0015
CSIDL_COMMON_STARTMENU        = 0x0016
CSIDL_COMMON_PROGRAMS         = 0X0017
CSIDL_COMMON_STARTUP          = 0x0018
CSIDL_COMMON_DESKTOPDIRECTORY = 0x0019
CSIDL_APPDATA                 = 0x001a
CSIDL_PRINTHOOD               = 0x001b
CSIDL_LOCAL_APPDATA           = 0x001c
CSIDL_ALTSTARTUP              = 0x001d
CSIDL_COMMON_ALTSTARTUP       = 0x001e
CSIDL_COMMON_FAVORITES        = 0x001f
CSIDL_INTERNET_CACHE          = 0x0020
CSIDL_COOKIES                 = 0x0021
CSIDL_HISTORY                 = 0x0022
CSIDL_COMMON_APPDATA          = 0x0023
CSIDL_WINDOWS                 = 0x0024
CSIDL_SYSTEM                  = 0x0025
CSIDL_PROGRAM_FILES           = 0x0026
CSIDL_MYPICTURES              = 0x0027
CSIDL_PROFILE                 = 0x0028
CSIDL_SYSTEMX86               = 0x0029
CSIDL_PROGRAM_FILESX86        = 0x002a
CSIDL_PROGRAM_FILES_COMMON    = 0x002b
CSIDL_PROGRAM_FILES_COMMONX86 = 0x002c
CSIDL_COMMON_TEMPLATES        = 0x002d
CSIDL_COMMON_DOCUMENTS        = 0x002e
CSIDL_COMMON_ADMINTOOLS       = 0x002f
CSIDL_ADMINTOOLS              = 0x0030
CSIDL_CONNECTIONS             = 0x0031
CSIDL_COMMON_MUSIC            = 0x0035
CSIDL_COMMON_PICTURES         = 0x0036
CSIDL_COMMON_VIDEO            = 0x0037
CSIDL_RESOURCES               = 0x0038
CSIDL_RESOURCES_LOCALIZED     = 0x0039
CSIDL_COMMON_OEM_LINKS        = 0x003a
CSIDL_CDBURN_AREA             = 0x003b
CSIDL_COMPUTERSNEARME         = 0x003d
CSIDL_FLAG_PER_USER_INIT      = 0x0800
CSIDL_FLAG_NO_ALIAS           = 0x1000
CSIDL_FLAG_DONT_UNEXPAND      = 0x2000
CSIDL_FLAG_DONT_VERIFY        = 0x4000
CSIDL_FLAG_CREATE             = 0x8000
CSIDL_FLAG_MASK               = 0xFF00

################################################################################

class SHELLEXECUTEINFOW(ctypes.Structure):
    _fields_ = (
        ("cbSize", DWORD),
        ("fMask", DWORD),
        ("hwnd", HWND),
        ("lpVerb", PWSTR),
        ("lpFile", PWSTR),
        ("lpParameters", PWSTR),
        ("lpDirectory", PWSTR),
        ("nShow", INT),
        ("hInstApp", HINSTANCE),
        ("lpIDList", PVOID),
        ("lpClass", PWSTR),
        ("hkeyClass", HANDLE),
        ("dwHotKey", DWORD),
        ("hMonitor", HANDLE),
        ("hProcess", HANDLE),
        )

    def __init__(self, file, verb, param, direc, wait, show):
        self.cbSize = ctypes.sizeof(self)
        self.lpVerb = verb
        self.lpFile = file
        self.lpParameters = param
        self.lpDirectory = direc
        self.nShow = show
        self.fMask = SEE_MASK_NOCLOSEPROCESS if wait else 0

PSHELLEXECUTEINFOW = POINTER(SHELLEXECUTEINFOW)

################################################################################

_ShellExecuteExW = fun_fact(_sh.ShellExecuteExW, (BOOL, PSHELLEXECUTEINFOW))

################################################################################

def ShellExecuteEx(
        file,
        verb=None,
        param=None,
        direc=None,
        wait=False,
        show=SW_SHOW
        ):
    sei = SHELLEXECUTEINFOW(str(file), verb, param, direc, wait, show)

    raise_on_zero(_ShellExecuteExW(ref(sei)))

    rc = None
    if sei.hProcess is not None:
        kernel.WaitForSingleObject(sei.hProcess, INFINITE)
        rc = kernel.GetExitCodeProcess(sei.hProcess)
        kernel.CloseHandle(sei.hProcess)
    return rc

################################################################################

def elevate(*args, direc=None, wait=False, show=SW_SHOW):
    file, param = args[0], cmdline_from_args(args[1:])
    ShellExecuteEx(file, "runas", param, direc, wait, show)

################################################################################

def relegate(*args, wait=False):
    _, pid, = user.GetWindowThreadProcessId(user.GetShellWindow())
    with kernel.OpenProcess(PROCESS_CREATE_PROCESS, False, pid) as shell_proc:
        attr = ((PROC_THREAD_ATTRIBUTE_PARENT_PROCESS, shell_proc),)
        with kernel.ProcThreadAttributeList(attr) as ptal:
            si = kernel.STARTUPINFOEX(ptal.address())
            cflags = CREATE_NEW_CONSOLE | EXTENDED_STARTUPINFO_PRESENT
            with kernel.create_process(args, cflags, si) as pi:
                if wait:
                    kernel.WaitForSingleObject(pi.hProcess, INFINITE)

################################################################################

_CommandLineToArgv = fun_fact(_sh.CommandLineToArgvW, (PPWSTR, PWSTR, PINT))

def CommandLineToArgv(cmdline):
    # CommandLineToArgv gets leading white space wrong
    cmdline = cmdline.lstrip(" \t")
    if not cmdline:
        return []

    argc = INT()
    pargs = _CommandLineToArgv(cmdline, ref(argc))
    raise_on_zero(pargs)
    try:
        return [pargs[i] for i in range(argc.value)]
    finally:
        kernel.LocalFree(pargs)

################################################################################

_SHGetFolderPath = fun_fact(
    _sh.SHGetFolderPathW, (HRESULT, HWND, INT, HANDLE, DWORD, PWSTR)
    )

def SHGetFolderPath(csidl, flags=0, token=None):
    buf = string_buffer(MAX_PATH)
    raise_on_hr(_SHGetFolderPath(None, csidl, token, flags, buf))
    return buf.value

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from enum import IntFlag as _int_flag
from types import SimpleNamespace as _namespace
from collections import defaultdict as _defdict

import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    BOOLEAN,
    BYTE,
    FILETIME,
    HANDLE,
    LARGE_INTEGER,
    LONG,
    LONG_PTR,
    INT,
    NTSTATUS,
    OSVERSIONINFOEX,
    PBOOLEAN,
    POINTER,
    POSVERSIONINFOEX,
    PULONG,
    PUNICODE_STRING,
    PVOID,
    SIZE_T,
    UINT_PTR,
    ULARGE_INTEGER,
    ULONG,
    ULONG_PTR,
    UNICODE_STRING,
    UnicodeStrBuffer,
    WCHAR,
    WCHAR_SIZE,
    WinError,
    WORD,
    )
from . import (
    ref,
    kernel,
    fun_fact,
    ns_from_struct,
    suppress_winerr,
    wtypes,
    ERROR_NO_MORE_FILES,
    DIRECTORY_QUERY,
    GENERIC_READ,
    SystemProcessInformation,
    SystemProcessIdInformation,
    SystemExtendedHandleInformation,
    ThreadBasicInformation,
    ThreadPriority,
    ProcessCommandLineInformation,
    ProcessImageFileName,
    ProcessBasicInformation,
    ProcessWow64Information,
    )

_nt = ctypes.WinDLL("ntdll.dll", use_last_error=True)

def _ntstatus(status):
    return LONG(status).value

################################################################################

RtlGetCurrentPeb = fun_fact(_nt.RtlGetCurrentPeb, (PVOID,))

################################################################################

STATUS_INFO_LENGTH_MISMATCH = _ntstatus(0xC0000004)
STATUS_BUFFER_OVERFLOW = _ntstatus(0x80000005)
STATUS_BUFFER_TOO_SMALL = _ntstatus(0xC0000023)
STATUS_INVALID_SIGNATURE = _ntstatus(0xC000A000)
STATUS_MORE_ENTRIES = _ntstatus(0x00000105)
STATUS_NO_MORE_ENTRIES= _ntstatus(0x8000001A)

################################################################################

_RtlNtStatusToDosError = fun_fact(
    _nt.RtlNtStatusToDosError, (ULONG, NTSTATUS)
    )

def RtlNtStatusToDosError(status):
    winerr = _RtlNtStatusToDosError(status)
    return winerr if winerr < 65536 else status

################################################################################

def raise_failed_status(status):
    if status < 0:
        raise WinError(RtlNtStatusToDosError(status))

################################################################################

class SYSTEM_PROCESS_ID_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("ProcessId", HANDLE),
        ("ImageName", UNICODE_STRING),
        )

################################################################################

class CLIENT_ID(ctypes.Structure):
    _fields_ = (
        ("UniqueProcess", LONG_PTR),
        ("UniqueThread", LONG_PTR)
        )

################################################################################

class SYSTEM_THREAD_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("KernelTime", LARGE_INTEGER),
        ("UserTime", LARGE_INTEGER),
        ("CreateTime", LARGE_INTEGER),
        ("WaitTime", ULONG),
        ("StartAddress", PVOID),
        ("ClientId", CLIENT_ID),
        ("Priority", LONG),
        ("BasePriority", LONG),
        ("ContextSwitches", ULONG),
        ("ThreadState", INT),
        ("WaitReason", INT),
        )

################################################################################

class SYSTEM_PROCESS_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("NextEntryOffset", ULONG),
        ("NumberOfThreads", ULONG),
        ("WorkingSetPrivateSize", LARGE_INTEGER),
        ("HardFaultCount", ULONG),
        ("NumberOfThreadsHighWatermark", ULONG),
        ("CycleTime", ULARGE_INTEGER),
        ("CreateTime", LARGE_INTEGER),
        ("UserTime", LARGE_INTEGER),
        ("KernelTime", LARGE_INTEGER),
        ("ImageName", UNICODE_STRING),  # file name only, no path
        ("BasePriority", LONG),
        ("UniqueProcessId", HANDLE),
        ("InheritedFromUniqueProcessId", HANDLE),
        ("HandleCount", ULONG),
        ("SessionId", ULONG),
        ("UniqueProcessKey", UINT_PTR),
        ("PeakVirtualSize", SIZE_T),
        ("VirtualSize", SIZE_T),
        ("PageFaultCount",  ULONG),
        ("PeakWorkingSetSize", SIZE_T),
        ("WorkingSetSize", SIZE_T),
        ("QuotaPeakPagedPoolUsage", SIZE_T),
        ("QuotaPagedPoolUsage", SIZE_T),
        ("QuotaPeakNonPagedPoolUsage", SIZE_T),
        ("QuotaNonPagedPoolUsage", SIZE_T),
        ("PagefileUsage", SIZE_T),
        ("PeakPagefileUsage", SIZE_T),
        ("PrivatePageCount", SIZE_T),
        ("ReadOperationCount", LARGE_INTEGER),
        ("WriteOperationCount", LARGE_INTEGER),
        ("OtherOperationCount", LARGE_INTEGER),
        ("ReadTransferCount", LARGE_INTEGER),
        ("WriteTransferCount", LARGE_INTEGER),
        ("OtherTransferCount", LARGE_INTEGER),
        )
        # At this point there is an array of
        # NumberOfThreads * SYSTEM_THREAD_INFORMATION.

################################################################################

class SYSTEM_BASIC_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("Reserved", ULONG),
        ("TimerResolution", ULONG),
        ("PageSize", ULONG),
        ("NumberOfPhysicalPages", ULONG),
        ("LowestPhysicalPageNumber", ULONG),
        ("HighestPhysicalPageNumber", ULONG),
        ("AllocationGranularity", ULONG),
        ("MinimumUserModeAddress", ULONG_PTR),
        ("MaximumUserModeAddress", ULONG_PTR),
        ("ActiveProcessorsAffinityMask", ULONG_PTR),
        ("NumberOfProcessors", BYTE),
        )

################################################################################

class SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("IdleTime", LARGE_INTEGER),
        ("KernelTime", LARGE_INTEGER),
        ("UserTime", LARGE_INTEGER),
        ("DpcTime", LARGE_INTEGER),
        ("InterruptTime", LARGE_INTEGER),
        ("InterruptCount", ULONG),
        )

################################################################################

class SYSTEM_HANDLE_TABLE_ENTRY_INFO(ctypes.Structure):
    _fields_ = (
        ("Object", PVOID),
        ("UniqueProcessId", UINT_PTR),
        ("HandleValue", UINT_PTR),
        ("GrantedAccess", ULONG),
        ("CreatorBackTraceIndex", WORD),
        ("ObjectTypeIndex", WORD),
        ("HandleAttributes", ULONG),
        ("Reserved", ULONG),
        )

################################################################################

class PROCESS_BASIC_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("ExitStatus", INT),
        ("PebBaseAddress", LONG_PTR),
        ("AffinityMask", LONG_PTR),
        ("BasePriority", INT),
        ("UniqueProcessId", LONG_PTR),
        ("InheritedFromUniqueProcessId", LONG_PTR)
        )

################################################################################

class PROCESS_EXTENDED_BASIC_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("Size", LONG_PTR),
        ("BasicInfo", PROCESS_BASIC_INFORMATION),
        ("Flags", INT)
        )

    def __init__(self):
        self.Size = ctypes.sizeof(self)

class PROCESS_EXTENDED_BASIC_FLAGS(_int_flag):
    IsProtectedProcess   = 1
    IsWow64Process       = 2
    IsProcessDeleting    = 4
    IsCrossSessionCreate = 8
    IsFrozen             = 16
    IsBackground         = 32
    IsStronglyNamed      = 64
    IsSecureProcess      = 128
    IsSubsystemProcess   = 256

################################################################################

class OBJECT_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ("Length", ULONG),
        ("RootDirectory", HANDLE),
        ("ObjectName", PUNICODE_STRING),
        ("Attributes", ULONG),
        ("SecurityDescriptor", PVOID),
        ("SecurityQualityOfService", PVOID)
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.Length = ctypes.sizeof(self)

def obj_attr(name, root=None, attr=0):
    name = wtypes.UnicodeStrBuffer(name).ptr if name else None
    return OBJECT_ATTRIBUTES(0, root, name, attr)

################################################################################

def _make_handle_info(num_entries):
    class _handle_info_t(ctypes.Structure):
        _fields_ = (
            ("NumberOfHandles", UINT_PTR),
            ("Reserved", UINT_PTR),
            ("Handles", SYSTEM_HANDLE_TABLE_ENTRY_INFO * num_entries)
            )
    return _handle_info_t()

################################################################################

_NtClose = fun_fact(_nt.NtClose, (NTSTATUS, PVOID))

def NtClose(handle):
    raise_failed_status(_NtClose(handle))

################################################################################

_NtOpenProcess = fun_fact(
    _nt.NtOpenProcess,
    (NTSTATUS, PVOID, ULONG, PVOID, PVOID)
    )

def NtOpenProcess(pid, desired_access):
    cid = CLIENT_ID()
    cid.UniqueProcess = pid
    h = HANDLE()
    oa = OBJECT_ATTRIBUTES()
    s = _NtOpenProcess(ref(h), desired_access, ref(oa), ref(cid))
    raise_failed_status(s)
    return h

################################################################################

_RtlAdjustPrivilege = fun_fact(
    _nt.RtlAdjustPrivilege,
    (NTSTATUS, ULONG, BOOLEAN, BOOLEAN, PBOOLEAN)
    )

def RtlAdjustPrivilege(priv, enable, thread_priv=False):
    prev = BOOLEAN()
    status = _RtlAdjustPrivilege(priv, enable, thread_priv, ref(prev))
    raise_failed_status(status)
    return bool(prev)

################################################################################

_NtQuerySystemInformation = fun_fact(
    _nt.NtQuerySystemInformation,
    (NTSTATUS, LONG, PVOID, ULONG, PULONG)
    )

def NtQuerySystemInformation(sys_info, buf, buf_size, p_ret_len):
    return _NtQuerySystemInformation(sys_info, buf, buf_size, p_ret_len)

################################################################################

_NtQueryInformationProcess = fun_fact(
    _nt.NtQueryInformationProcess,
    (NTSTATUS, PVOID, LONG, PVOID, ULONG, PVOID)
    )

def NtQueryInformationProcess(phandle, pinfo, buf, buf_size, p_ret_len):
    return _NtQueryInformationProcess(phandle, pinfo, buf, buf_size, p_ret_len)

################################################################################

def _var_size_proc_info(proc_handle, proc_info):
    # This works only for the few information classes that return variable
    # size items (50, 51, 60, 64, 85, 97). Those that deliver fixed size items
    # cannot be queried for the required size.
    size = ULONG(0)
    NtQueryInformationProcess(proc_handle, proc_info, 0, size, ref(size))
    buff = byte_buffer(size.value)
    raise_failed_status(
        NtQueryInformationProcess(proc_handle, proc_info, buff, size, ref(size))
        )
    return buff

################################################################################

def _fixed_size_proc_info(proc_handle, proc_info, dest):
    size = ctypes.sizeof(dest)
    raise_failed_status(
        NtQueryInformationProcess(proc_handle, proc_info, ref(dest), size, None)
        )

################################################################################

def _enum_proc_worker(extract_func):
    status = STATUS_INFO_LENGTH_MISMATCH
    while status == STATUS_INFO_LENGTH_MISMATCH:
        size = ULONG(0)
        NtQuerySystemInformation(SystemProcessInformation, 0, 0, ref(size))
        buf = byte_buffer(size.value)
        status = NtQuerySystemInformation(
            SystemProcessInformation,
            ref(buf),
            size,
            ref(size)
            )
    raise_failed_status(status)

    res = []
    pinfo_size = ctypes.sizeof(SYSTEM_PROCESS_INFORMATION)
    offs, offs_inc = 0, 1
    while offs_inc:
        pi = SYSTEM_PROCESS_INFORMATION.from_buffer(buf, offs)
        nt = pi.NumberOfThreads
        ti_addr = ctypes.addressof(pi) + pinfo_size
        ti = (SYSTEM_THREAD_INFORMATION * nt).from_address(ti_addr)
        res.append(extract_func(pi, ti))
        offs_inc = pi.NextEntryOffset
        offs += offs_inc
    return res

################################################################################

def _extract_basic_proc_info(pi, ti):
    return _namespace(
        name=(
            str(pi.ImageName) if pi.ImageName.Buffer else
            ("idle" if not pi.UniqueProcessId else "system")
            ),
        pid=pi.UniqueProcessId or 0,
        tids=[i.ClientId.UniqueThread for i in ti],
        )

################################################################################

def _extract_full_proc_info(pi, ti):
    pins = ns_from_struct(pi)
    del pins.NextEntryOffset
    pins.Threads = [ns_from_struct(t) for t in ti]
    return pins

################################################################################

def enum_processes():
    return _enum_proc_worker(_extract_basic_proc_info)

################################################################################

def enum_processes_ex():
    return _enum_proc_worker(_extract_full_proc_info)

################################################################################

def _resolve_device_prefix(fname):
    gattr = obj_attr("\\GLOBAL??")
    with NtOpenDirectoryObject(DIRECTORY_QUERY, gattr) as glob:
        for prefix, typ in NtQueryDirectoryObject(glob):
            if len(prefix) != 2 or prefix[1] != ":" or typ != "SymbolicLink":
                continue
            lattr = obj_attr(prefix, glob)
            with NtOpenSymbolicLinkObject(GENERIC_READ, lattr) as hsl:
                link = NtQuerySymbolicLinkObject(hsl)
                if fname.startswith(link):
                    return fname.replace(link, prefix, 1)
    return fname

################################################################################

def proc_path_from_pid(pid):

    if pid == 0:
        return "idle"
    elif pid == 4:
        return "system"

    buf = string_buffer(512)
    size = ctypes.sizeof(SYSTEM_PROCESS_ID_INFORMATION)
    spii = SYSTEM_PROCESS_ID_INFORMATION()
    spii.ProcessId = ctypes.cast(pid, HANDLE)
    spii.ImageName.Length = 0
    spii.ImageName.MaximumLength = buf._length_
    spii.ImageName.Buffer = ctypes.addressof(buf)

    while True:
        status = NtQuerySystemInformation(
            SystemProcessIdInformation,
            ref(spii),
            size,
            None
            )
        if status != STATUS_INFO_LENGTH_MISMATCH:
            break

        # Required length is stored in MaximumLength.
        buf = string_buffer(spii.ImageName.MaximumLength)
        spii.ImageName.Buffer = ctypes.addressof(buf)

    raise_failed_status(status)
    return _resolve_device_prefix(str(spii.ImageName))

################################################################################

def proc_path_from_handle(handle):
    buf = _var_size_proc_info(handle, ProcessImageFileName)
    return _resolve_device_prefix(str(UNICODE_STRING.from_buffer(buf)))

################################################################################

def get_handles(pid=-1):
    info = SystemExtendedHandleInformation
    hi = _make_handle_info(1)
    rlen = ULONG(0)
    while True:
        _NtQuerySystemInformation(info, ref(hi), ctypes.sizeof(hi), ref(rlen))
        hi = _make_handle_info(hi.NumberOfHandles)
        status = _NtQuerySystemInformation(
            info,
            ref(hi),
            ctypes.sizeof(hi),
            ref(rlen)
            )
        if status == STATUS_INFO_LENGTH_MISMATCH:
            continue
        raise_failed_status(status)
        if pid == -1:
            return list(hi.Handles)
        return [h for h in hi.Handles if pid == h.UniqueProcessId]

################################################################################

def get_grouped_handles(pid=-1):
    grouped_handles = _defdict(list)
    for h in get_handles(pid):
        grouped_handles[h.UniqueProcessId].append(h)
    return grouped_handles

################################################################################

_NtGetNextProcess = fun_fact(
    _nt.NtGetNextProcess,
    (NTSTATUS, PVOID, ULONG, ULONG, ULONG, PVOID)
    )

def NtGetNextProcess(current, access, attribs=0, flags=0):
    nxt = HANDLE()
    # have to ignore returned NTSTATUS, success/failure is conveyed by the
    # nxt handle
    _NtGetNextProcess(current, access, attribs, flags, ref(nxt))
    return nxt.value

################################################################################

def get_proc_ext_basic_info(proc_handle):
    pebi = PROCESS_EXTENDED_BASIC_INFORMATION()
    _fixed_size_proc_info(proc_handle, ProcessBasicInformation, pebi)
    return pebi

################################################################################

def pid_from_handle(proc_handle):
    return get_proc_ext_basic_info(proc_handle).BasicInfo.UniqueProcessId

################################################################################

class _DUMMY_STATUS_UNION(ctypes.Union):
    _fields_ = (
        ("Status", LONG),
        ("Pointer", PVOID),
        )

class IO_STATUS_BLOCK(ctypes.Structure):
    _anonymous_ = ("anon",)
    _fields_ = (
        ("anon", _DUMMY_STATUS_UNION),
        ("Information", UINT_PTR),
        )
PIO_STATUS_BLOCK = POINTER(IO_STATUS_BLOCK)

################################################################################

class FILE_DIRECTORY_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("NextEntryOffset", ULONG),
        ("FileIndex", ULONG),
        ("CreationTime", LARGE_INTEGER),
        ("LastAccessTime", LARGE_INTEGER),
        ("LastWriteTime", LARGE_INTEGER),
        ("ChangeTime", LARGE_INTEGER),
        ("EndOfFile", LARGE_INTEGER),
        ("AllocationSize", LARGE_INTEGER),
        ("FileAttributes", ULONG),
        ("FileNameLength", ULONG),
        ("FileName", WCHAR * 1),
        )

PFILE_DIRECTORY_INFORMATION = POINTER(FILE_DIRECTORY_INFORMATION)

################################################################################

_NtQueryDirectoryFile = fun_fact(
    _nt.NtQueryDirectoryFile, (
        NTSTATUS,
        HANDLE,
        HANDLE,
        PVOID,  # PIO_APC_ROUTINE
        PVOID,
        PIO_STATUS_BLOCK,
        PVOID,
        ULONG,
        INT,
        BOOLEAN,
        PUNICODE_STRING,
        BOOLEAN,
        )
    )

################################################################################

def get_directory_info(hdir, restart_scan):
    iosb = IO_STATUS_BLOCK()
    bsize = 1024
    while True:
        buf = byte_buffer(bsize)
        stat = _NtQueryDirectoryFile(
            hdir,
            None,
            None,
            None,
            ref(iosb),
            buf,
            bsize,
            1,  # FileDirectoryInformation
            False,
            None,
            restart_scan
            )
        if stat in (STATUS_BUFFER_OVERFLOW, STATUS_INFO_LENGTH_MISMATCH):
            bsize *= 2
        else:
            break
    raise_failed_status(stat)

    def extract_info(addr):
        def la2dt(la):
            return kernel.FileTimeToLocalSystemTime(
                FILETIME(la)
                ).to_datetime()

        info = FILE_DIRECTORY_INFORMATION.from_address(addr)
        name = ctypes.wstring_at(
            addr + FILE_DIRECTORY_INFORMATION.FileName.offset,
            info.FileNameLength // WCHAR_SIZE
            )
        nxt = addr + info.NextEntryOffset if info.NextEntryOffset else 0
        return nxt, _namespace(
            FileIndex=info.FileIndex,
            CreationTime=la2dt(info.CreationTime),
            LastAccessTime=la2dt(info.LastAccessTime),
            LastWriteTime=la2dt(info.LastWriteTime),
            ChangeTime=la2dt(info.ChangeTime),
            EndOfFile=info.EndOfFile,
            AllocationSize=info.AllocationSize,
            FileAttributes=info.FileAttributes,
            FileName=name,
            )

    addr = ctypes.addressof(buf)
    while addr:
        addr, info = extract_info(addr)
        yield info

################################################################################

def enum_directory_info(hdir):
    restart_scan = True
    with suppress_winerr(ERROR_NO_MORE_FILES):
        while True:
            yield from get_directory_info(hdir, restart_scan)
            restart_scan = False

################################################################################

_RtlGetVersion = fun_fact(_nt.RtlGetVersion, (NTSTATUS, POSVERSIONINFOEX))

def RtlGetVersion():
    osve = OSVERSIONINFOEX()
    raise_failed_status(_RtlGetVersion(ref(osve)))
    return ns_from_struct(osve)

################################################################################

_NtPowerInformation = fun_fact(
    _nt.NtPowerInformation, (
        NTSTATUS,
        LONG,
        PVOID,
        ULONG,
        PVOID,
        ULONG,
        )
    )

def NtPowerInformation(pwr_info, in_bytes, out_len):
    if in_bytes is None:
        iptr, ilen = None, 0
    else:
        iptr, ilen = ref(in_bytes), len(in_bytes)

    if out_len is None or out_len == 0:
        out, optr, olen = byte_buffer(0), None, 0
    else:
        out = byte_buffer(out_len)
        optr, olen = ref(out), out_len

    raise_failed_status(_NtPowerInformation(pwr_info, iptr, ilen, optr, olen))
    return out

################################################################################

def get_proc_command_line(proc_handle):
    buf = _var_size_proc_info(proc_handle, ProcessCommandLineInformation)
    return str(UNICODE_STRING.from_buffer(buf))

################################################################################

def get_proc_env_blk(proc_handle):
    return get_proc_ext_basic_info(proc_handle).BasicInfo.PebBaseAddress

################################################################################

def get_wow64_proc_env_blk(proc_handle):
    peb32 = ULONG_PTR()
    _fixed_size_proc_info(proc_handle, ProcessWow64Information, peb32)
    return peb32.value

################################################################################

class THREAD_BASIC_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("ExitStatus", LONG),
        ("TebBaseAddress", PVOID),
        ("ClientId", CLIENT_ID),
        ("AffinityMask", ULONG_PTR),
        ("Priority", LONG),
        ("BasePriority", LONG),
        )

################################################################################

_NtQueryInformationThread = fun_fact(
    _nt.NtQueryInformationThread,
    (NTSTATUS, PVOID, LONG, PVOID, ULONG, PVOID)
    )

def NtQueryInformationThread(thdl, tinfo, buf, buf_size, p_ret_len):
    return _NtQueryInformationThread(thdl, tinfo, buf, buf_size, p_ret_len)

################################################################################

def get_thread_basic_info(thdl):
    tbi = THREAD_BASIC_INFORMATION()
    size = ctypes.sizeof(tbi)
    raise_failed_status(
        NtQueryInformationThread(
            thdl,
            ThreadBasicInformation,
            ref(tbi),
            size,
            None
            )
        )
    return tbi

################################################################################

_NtSetInformationThread = fun_fact(
    _nt.NtSetInformationThread,
    (NTSTATUS, PVOID, LONG, PVOID, ULONG)
    )

def NtSetInformationThread(thdl, tinfo, ct_info_obj):
    raise_failed_status(
        _NtSetInformationThread(
            thdl,
            tinfo,
            ref(ct_info_obj),
            ctypes.sizeof(ct_info_obj)
            )
        )

################################################################################

# requires SeIncreaseBasePriorityPrivilege

def set_abs_thread_priority(thdl, prio):
    prio = ULONG(prio)
    NtSetInformationThread(thdl, ThreadPriority, prio)

################################################################################

def get_abs_thread_priority(thdl):
    tbi = get_thread_basic_info(thdl)
    return tbi.Priority

################################################################################

_NtGetNextThread = fun_fact(
    _nt.NtGetNextThread,
    (NTSTATUS, PVOID, PVOID, ULONG, ULONG, ULONG, PVOID)
    )

def NtGetNextThread(proc, cur_thrd, access, attribs=0, flags=0):
    nxt = HANDLE()
    _NtGetNextThread(proc, cur_thrd, access, attribs, flags, ref(nxt))
    return nxt.value

################################################################################

_NtOpenDirectoryObject = fun_fact(
    _nt.NtOpenDirectoryObject,
    (NTSTATUS, PVOID, ULONG, PVOID)
    )

def NtOpenDirectoryObject(acc, obj_attr):
    hdl = kernel.KHANDLE()
    raise_failed_status(_NtOpenDirectoryObject(ref(hdl), acc, ref(obj_attr)))
    return hdl

################################################################################

class OBJECT_DIRECTORY_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("Name", UNICODE_STRING),
        ("TypeName", UNICODE_STRING),
        )

################################################################################

_NtQueryDirectoryObject = fun_fact(
    _nt.NtQueryDirectoryObject,
    (NTSTATUS, HANDLE, PVOID, ULONG, BOOLEAN, BOOLEAN, PULONG, PULONG)
    )

def NtQueryDirectoryObject(hdir):
    res = []
    bsize = 4096
    buf = byte_buffer(bsize)
    context = ULONG()
    rlen = ULONG()
    restart = True
    while True:
        stat = _NtQueryDirectoryObject(
            hdir,
            buf,
            bsize,
            False,
            restart,
            ref(context),
            ref(rlen)
            )
        if stat >= 0:
            addr = ctypes.addressof(buf)
            while True:
                info = OBJECT_DIRECTORY_INFORMATION.from_address(addr)
                if info.Name.Length == 0:
                    break
                res.append((str(info.Name), str(info.TypeName)))
                addr += ctypes.sizeof(OBJECT_DIRECTORY_INFORMATION)
            restart = False
            if stat != STATUS_MORE_ENTRIES:
                break
        elif stat == STATUS_NO_MORE_ENTRIES:
            break
        else:
            raise_failed_status(stat)
    return res

################################################################################

_NtOpenSymbolicLinkObject = fun_fact(
    _nt.NtOpenSymbolicLinkObject,
    (NTSTATUS, PVOID, ULONG, PVOID)
    )

def NtOpenSymbolicLinkObject(acc, obj_attr):
    hdl = kernel.KHANDLE()
    raise_failed_status(_NtOpenSymbolicLinkObject(ref(hdl), acc, ref(obj_attr)))
    return hdl

################################################################################

_NtQuerySymbolicLinkObject = fun_fact(
    _nt.NtQuerySymbolicLinkObject,
    (NTSTATUS, HANDLE, PUNICODE_STRING, PULONG)
    )

def NtQuerySymbolicLinkObject(hdl):
    size = 256
    us = UnicodeStrBuffer(size)
    rlen = ULONG()
    while True:
        stat = _NtQuerySymbolicLinkObject(hdl, us.ptr, ref(rlen))
        if stat == STATUS_BUFFER_TOO_SMALL:
            size *= 2
            us = UnicodeStrBuffer(size)
        else:
            raise_failed_status(stat)
            return us.str

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from ctwin32 import (
    ctypes,
    user,
    gdi,
    wndcls,
    comctl,
    shell,
    WM_CONTEXTMENU,
    WM_CREATE,
    WM_PAINT,
    WM_DESTROY,
    DT_CENTER,
    DT_SINGLELINE,
    DT_VCENTER,
    TDN_HYPERLINK_CLICKED,
    TDF_ALLOW_DIALOG_CANCELLATION,
    TDF_ENABLE_HYPERLINKS,
    TDF_POSITION_RELATIVE_TO_WINDOW,
    TDCBF_CLOSE_BUTTON,
    TD_INFORMATION_ICON,
    )
from ctwin32.wtypes import LOGFONT

################################################################################

def td_follow_link(hwnd, msg, wp, lp, ctxt):
    if msg == TDN_HYPERLINK_CLICKED:
        shell.ShellExecuteEx(ctypes.wstring_at(lp))

################################################################################

class HelloWnd(wndcls.SimpleWnd):

    def on_message(self, msg, wp, lp):

        if msg == WM_CREATE:
            lf = LOGFONT()
            lf.lfFaceName = "MS Shell Dlg"
            lf.lfHeight = -72
            self.font = gdi.CreateFontIndirect(lf)
            return 0

        elif msg == WM_PAINT:
            hdc, ps = self.begin_paint()
            oldfont = gdi.SelectObject(hdc, self.font)
            user.DrawText(
                hdc,
                "Hello from ctwin32!",
                self.client_rect(),
                DT_CENTER | DT_SINGLELINE | DT_VCENTER
                )
            gdi.SelectObject(hdc, oldfont)
            self.end_paint(ps)
            return 0

        elif msg == WM_DESTROY:
            gdi.DeleteObject(self.font)
            user.PostQuitMessage(0)
            return 0

        elif msg == WM_CONTEXTMENU:
            tdi = "TaskDialogIndirect"
            url = f"https://www.google.com/search?q={tdi}"
            tdc = comctl.TASKDIALOGCONFIG()
            tdc.hwndParent = self.hwnd
            tdc.dwFlags = (
                TDF_ALLOW_DIALOG_CANCELLATION |
                TDF_ENABLE_HYPERLINKS |
                TDF_POSITION_RELATIVE_TO_WINDOW
                )
            tdc.dwCommonButtons = TDCBF_CLOSE_BUTTON
            tdc.pszWindowTitle = "TaskDialog Sample"
            tdc.pszMainIcon = TD_INFORMATION_ICON
            tdc.pszMainInstruction = f"Showing off {tdi}!"
            tdc.pszContent = f'Search Google for <a href="{url}">{tdi}</a>.'
            comctl.tsk_dlg_callback(tdc, td_follow_link)
            return 0

        else:
            return self.def_win_proc(msg, wp, lp)

################################################################################

if __name__ == "__main__":

    icon = wndcls.load_ctwin32_ico()
    wnd = HelloWnd(wndcls.WndCreateParams("Hello Window", icon))
    wnd.show()

    while msg := user.GetMessage():
        user.TranslateMessage(msg)
        user.DispatchMessage(msg)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from ctwin32.bcrypt import (
    BCRYPT_AES_ALGORITHM,
    BCRYPT_CHAINING_MODE,
    BCRYPT_CHAIN_MODE_CBC,
    BCryptOpenAlgorithmProvider,
    BCryptSetProperty,
    BCryptGenerateSymmetricKey,
    BCryptEncrypt,
    BCryptDecrypt,
    )

################################################################################

CRYPT_BLOCK_LEN = 16
IS_PYAES = False
try:
    import pyaes
    IS_PYAES = True
except ImportError:
    from cryptography.hazmat.primitives.ciphers import (
        Cipher,
        algorithms,
        modes
        )

################################################################################

def simple_py_aes_encrypt(data, key):
    blen = CRYPT_BLOCK_LEN
    if IS_PYAES:
        rng = range(0, len(data), blen)
        aes = pyaes.AESModeOfOperationCBC(key, None)
        return b"".join(aes.encrypt(data[i:i + blen]) for i in rng)
    else:
        aes = Cipher(algorithms.AES(key), modes.CBC(b"\0" * blen))
        encryptor = aes.encryptor()
        return encryptor.update(data) + encryptor.finalize()

################################################################################

def simple_py_aes_decrypt(data, key):
    blen = CRYPT_BLOCK_LEN
    if IS_PYAES:
        rng = range(0, len(data), blen)
        aes = pyaes.AESModeOfOperationCBC(key, None)
        return b"".join(aes.decrypt(data[i:i + blen]) for i in rng)
    else:
        aes = Cipher(algorithms.AES(key), modes.CBC(b"\0" * blen))
        decryptor = aes.decryptor()
        return decryptor.update(data) + decryptor.finalize()

################################################################################

def simple_bc_aes_encrypt(data, key):
    with BCryptOpenAlgorithmProvider(BCRYPT_AES_ALGORITHM) as alg:
        BCryptSetProperty(alg, BCRYPT_CHAINING_MODE, BCRYPT_CHAIN_MODE_CBC)
        with BCryptGenerateSymmetricKey(alg, key) as hkey:
            return BCryptEncrypt(hkey, data)

################################################################################

def simple_bc_aes_decrypt(data, key):
    with BCryptOpenAlgorithmProvider(BCRYPT_AES_ALGORITHM) as alg:
        BCryptSetProperty(alg, BCRYPT_CHAINING_MODE, BCRYPT_CHAIN_MODE_CBC)
        with BCryptGenerateSymmetricKey(alg, key) as hkey:
            return BCryptDecrypt(hkey, data)

################################################################################

_ascii_trans = bytes([i if 32 <= i < 128 else ord(".") for i in range(256)])

################################################################################

def hex_dump(data, bytes_per_line=16):
    lines = []
    hex_chars = bytes_per_line * 3 - 1
    length = len(data)
    offset = 0
    while offset < length:
        chunk_len = min(length - offset, bytes_per_line)
        chunk = data[offset:offset + chunk_len]
        ascii = chunk.translate(_ascii_trans).decode("ascii")
        hexa = " ".join([f"{c:02x}" for c in chunk]).ljust(hex_chars)
        lines.append(f"{offset:08x} | {hexa} | {ascii}")
        offset += chunk_len
    return "\n".join(lines)

################################################################################

if __name__ == "__main__":

    try:
        assert 1 == 2 # noqa: PLR0133
    except AssertionError:
        pass
    else:
        raise RuntimeError("Assertions are not active")

    plain = bytes(range(256))
    key = b"\xf8\xa9V\x9e'#1\xd7\x1a\x85\r)\x02\x8c,\xa9"

    simple_py_aes_enc = simple_py_aes_encrypt(plain, key)
    assert simple_py_aes_decrypt(simple_py_aes_enc, key) == plain

    simple_bc_aes_enc = simple_bc_aes_encrypt(plain, key)
    assert simple_bc_aes_decrypt(simple_bc_aes_enc, key) == plain

    assert simple_py_aes_decrypt(simple_bc_aes_enc, key) == plain
    assert simple_bc_aes_decrypt(simple_py_aes_enc, key) == plain

    assert simple_py_aes_enc == simple_bc_aes_enc

    print(f"\nplain:\n{hex_dump(plain)}")
    print(f"\nencrypted:\n{hex_dump(simple_bc_aes_enc)}")



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# inspired by
# https://randomascii.wordpress.com/2018/02/11/zombie-processes-are-eating-your-memory/

import sys
import collections
from ctwin32 import (
    ntdll,
    kernel,
    PROCESS_QUERY_LIMITED_INFORMATION,
    SE_DEBUG_PRIVILEGE
    )

################################################################################

def get_zombies(verbose):
    zombies = {}
    is_deleting = ntdll.PROCESS_EXTENDED_BASIC_FLAGS.IsProcessDeleting
    add_lf = False

    hdl = ntdll.NtGetNextProcess(None, PROCESS_QUERY_LIMITED_INFORMATION)
    while hdl:
        close_me = hdl
        try:
            pebi = ntdll.get_proc_ext_basic_info(hdl)
        except OSError as e:
            print(e)
        else:
            if (pebi.Flags & is_deleting) != 0:
                path = ntdll.proc_path_from_handle(hdl)
                zombies[hdl] = path
                close_me = None
                if verbose:
                    zpid = pebi.BasicInfo.UniqueProcessId
                    ppid = pebi.BasicInfo.InheritedFromUniqueProcessId
                    print(f"zpid = {zpid:5}, ppid = {ppid:5}, {path}")
                    add_lf = True
        hdl = ntdll.NtGetNextProcess(hdl, PROCESS_QUERY_LIMITED_INFORMATION)
        # only keep handles of zombies open
        if close_me:
            ntdll.NtClose(close_me)
    if add_lf:
        print()
    return zombies

################################################################################

def get_zombie_object_addresses(verbose):
    z = get_zombies(verbose)
    handles = ntdll.get_handles(kernel.GetCurrentProcessId())
    res = {h.Object: z[h.HandleValue] for h in handles if h.HandleValue in z}
    # Close the handles to the zombies so that we do not report that we ourself
    # hold on to them.
    for h in z:
        ntdll.NtClose(h)
    return res

################################################################################

def main():

    try:
        ntdll.RtlAdjustPrivilege(SE_DEBUG_PRIVILEGE, True)
    except OSError:
        print("\nFailed to enable privilege. Some results may be missing.\n")

    verbose = "-v" in sys.argv or "--verbose" in sys.argv

    def cnt2str(cnt):
        return f"{cnt} zombie{'s' if cnt > 1 else ''}"

    zombies = get_zombie_object_addresses(verbose)
    if len(zombies):
        print(f"Found {cnt2str(len(zombies))}.")
    else:
        print("No zombies found.")

    pids_and_paths = []
    gh = ntdll.get_grouped_handles()
    for pid, infos in gh.items():
        paths = []
        for info in infos:
            path = zombies.get(info.Object, "")
            if path:
                paths.append(path)
                del zombies[info.Object]
        if paths:
            pids_and_paths.append((pid, paths))

    for pid, zpaths in pids_and_paths:
        hpath = ntdll.proc_path_from_pid(pid)
        print(f"\n{cnt2str(len(zpaths))} held by {hpath} ({pid})")
        counted_zombies = collections.defaultdict(int)
        for p in zpaths:
            counted_zombies[p] += 1
        for path, cnt in counted_zombies.items():
            print(f"    {cnt2str(cnt)} of {path}")

    if zombies:
        print("\nFound no owner for:")
        for z in zombies.values():
            print(f"    {z}")

################################################################################

if __name__ == "__main__":
    main()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from . import (
    kernel,
    ntdll,
    ERROR_BAD_EXE_FORMAT,
    FILE_MAP_READ,
    GENERIC_READ,
    IMAGE_DOS_SIGNATURE,
    IMAGE_NT_SIGNATURE,
    IMAGE_NT_OPTIONAL_HDR64_MAGIC,
    IMAGE_NUMBEROF_DIRECTORIES,
    IMAGE_SIZEOF_SHORT_NAME,
    PAGE_READONLY,
    )
from .wtypes import (
    BYTE,
    DWORD,
    PTR_HAS_64_BITS,
    PVOID,
    ULONG,
    ULONGLONG,
    WCHAR_SIZE,
    WORD,
    )

################################################################################

class classproperty(property):
    def __get__(self, owner, owner_type):
        return self.fget(owner_type)

################################################################################

class SizedStruct(ctypes.Structure):
    @classproperty
    def _size_(cls): # noqa: N805 this IS a class method
        return ctypes.sizeof(cls)

################################################################################

class IMAGE_FILE_HEADER(SizedStruct):
    _fields_ = (
        ("Machine", WORD),
        ("NumberOfSections", WORD),
        ("TimeDateStamp", DWORD),
        ("PointerToSymbolTable", DWORD),
        ("NumberOfSymbols", DWORD),
        ("SizeOfOptionalHeader", WORD),
        ("Characteristics", WORD),
        )

################################################################################

class IMAGE_DATA_DIRECTORY(SizedStruct):
    _fields_ = (
        ("VirtualAddress", DWORD),
        ("Size", DWORD),
        )

################################################################################

class IMAGE_OPTIONAL_HEADER32(SizedStruct):
    _fields_ = (
        ("Magic", WORD),
        ("MajorLinkerVersion", BYTE),
        ("MinorLinkerVersion", BYTE),
        ("SizeOfCode", DWORD),
        ("SizeOfInitializedData", DWORD),
        ("SizeOfUninitializedData", DWORD),
        ("AddressOfEntryPoint", DWORD),
        ("BaseOfCode", DWORD),
        ("BaseOfData", DWORD),
        ("ImageBase", DWORD),
        ("SectionAlignment", DWORD),
        ("FileAlignment", DWORD),
        ("MajorOperatingSystemVersion", WORD),
        ("MinorOperatingSystemVersion", WORD),
        ("MajorImageVersion", WORD),
        ("MinorImageVersion", WORD),
        ("MajorSubsystemVersion", WORD),
        ("MinorSubsystemVersion", WORD),
        ("Win32VersionValue", DWORD),
        ("SizeOfImage", DWORD),
        ("SizeOfHeaders", DWORD),
        ("CheckSum", DWORD),
        ("Subsystem", WORD),
        ("DllCharacteristics", WORD),
        ("SizeOfStackReserve", DWORD),
        ("SizeOfStackCommit", DWORD),
        ("SizeOfHeapReserve", DWORD),
        ("SizeOfHeapCommit", DWORD),
        ("LoaderFlags", DWORD),
        ("NumberOfRvaAndSizes", DWORD),
        ("DataDirectory", IMAGE_DATA_DIRECTORY * IMAGE_NUMBEROF_DIRECTORIES),
        )

################################################################################

class IMAGE_OPTIONAL_HEADER64(SizedStruct):
    _fields_ = (
        ("Magic", WORD),
        ("MajorLinkerVersion", BYTE),
        ("MinorLinkerVersion", BYTE),
        ("SizeOfCode", DWORD),
        ("SizeOfInitializedData", DWORD),
        ("SizeOfUninitializedData", DWORD),
        ("AddressOfEntryPoint", DWORD),
        ("BaseOfCode", DWORD),
        ("ImageBase", ULONGLONG),
        ("SectionAlignment", DWORD),
        ("FileAlignment", DWORD),
        ("MajorOperatingSystemVersion", WORD),
        ("MinorOperatingSystemVersion", WORD),
        ("MajorImageVersion", WORD),
        ("MinorImageVersion", WORD),
        ("MajorSubsystemVersion", WORD),
        ("MinorSubsystemVersion", WORD),
        ("Win32VersionValue", DWORD),
        ("SizeOfImage", DWORD),
        ("SizeOfHeaders", DWORD),
        ("CheckSum", DWORD),
        ("Subsystem", WORD),
        ("DllCharacteristics", WORD),
        ("SizeOfStackReserve", ULONGLONG),
        ("SizeOfStackCommit", ULONGLONG),
        ("SizeOfHeapReserve", ULONGLONG),
        ("SizeOfHeapCommit", ULONGLONG),
        ("LoaderFlags", DWORD),
        ("NumberOfRvaAndSizes", DWORD),
        ("DataDirectory", IMAGE_DATA_DIRECTORY * IMAGE_NUMBEROF_DIRECTORIES),
        )

################################################################################

class IMAGE_SECTION_HEADER(SizedStruct):
    _fields_ = (
        ("Name", BYTE * IMAGE_SIZEOF_SHORT_NAME),
        ("VirtualSize", DWORD),
        ("VirtualAddress", DWORD),
        ("SizeOfRawData", DWORD),
        ("PointerToRawData", DWORD),
        ("PointerToRelocations", DWORD),
        ("PointerToLinenumbers", DWORD),
        ("NumberOfRelocations", WORD),
        ("NumberOfLinenumbers", WORD),
        ("Characteristics", DWORD),
        )

################################################################################

class IMAGE_EXPORT_DIRECTORY(SizedStruct):
    _fields_ = (
        ("Characteristics", DWORD),
        ("TimeDateStamp", DWORD),
        ("MajorVersion", WORD),
        ("MinorVersion", WORD),
        ("Name", DWORD),
        ("Base", DWORD),
        ("NumberOfFunctions", DWORD),
        ("NumberOfNames", DWORD),
        ("AddressOfFunctions", DWORD),
        ("AddressOfNames", DWORD),
        ("AddressOfNameOrdinals", DWORD),
        )

################################################################################

class IID_UNION(ctypes.Union):
    _fields_ = (
        ("Characteristics", DWORD),
        ("OriginalFirstThunk", DWORD),
        )

class IMAGE_IMPORT_DESCRIPTOR(SizedStruct):
    _anonymous_ = ("anon",)
    _fields_ = (
        ("anon", IID_UNION),
        ("TimeDateStamp", DWORD),
        ("ForwarderChain", DWORD),
        ("Name", DWORD),
        ("FirstThunk", DWORD),
        )

################################################################################

class IMAGE_DELAYLOAD_DESCRIPTOR(SizedStruct):
    _fields_ = (
        ("Attributes", DWORD),
        ("DllNameRVA", DWORD),
        ("ModuleHandleRVA", DWORD),
        ("ImportAddressTableRVA", DWORD),
        ("ImportNameTableRVA", DWORD),
        ("BoundImportAddressTableRVA", DWORD),
        ("UnloadInformationTableRVA", DWORD),
        ("TimeDateStamp", DWORD),
        )

dlattrRva = 0x1

################################################################################

class NotPeError(OSError):
    def __init__(self, name, code=None):
        code = code or ERROR_BAD_EXE_FORMAT
        super().__init__(0, "not a PE file", str(name), code)

################################################################################

class pemap:
    E_MAGIC = 0     # file offset of 'MZ' magic
    E_LFANEW = 60   # file offset of NT header offset

    def __init__(self, fname, *, convert_open_err=False):
        try:
            f = kernel.create_file(str(fname), GENERIC_READ)
        except OSError as e:
            if convert_open_err:
                raise NotPeError(fname, e.winerror) from e
            raise

        with f:
            self.size = kernel.GetFileSize(f)
            if self.size < (self.E_LFANEW + 4):
                raise NotPeError(fname)
            with kernel.CreateFileMapping(f, None, PAGE_READONLY, 0) as m:
                self.view = kernel.MapViewOfFile(m, FILE_MAP_READ, 0, 0)

        if fhaddr := self._file_header_addr():
            self.name = fname
            self.key = str(fname).split("\\")[-1].lower()
            self.file_hdr = IMAGE_FILE_HEADER.from_address(fhaddr)
            ohaddr = fhaddr + ctypes.sizeof(IMAGE_FILE_HEADER)
            self.opt_hdr = IMAGE_OPTIONAL_HEADER64.from_address(ohaddr)
            self.is64bit = True
            if self.opt_hdr.Magic != IMAGE_NT_OPTIONAL_HDR64_MAGIC:
                self.is64bit = False
                self.opt_hdr = IMAGE_OPTIONAL_HEADER32.from_address(ohaddr)

            secaddr = ohaddr + self.file_hdr.SizeOfOptionalHeader
            size = ctypes.sizeof(IMAGE_SECTION_HEADER)
            self.sections = [
                IMAGE_SECTION_HEADER.from_address(secaddr + i * size)
                for i in range(self.file_hdr.NumberOfSections)
                ]
        else:
            self.close()
            raise NotPeError(fname)

    ############################################################################

    def close(self):
        kernel.UnmapViewOfFile(self.view)
        self.view = self.file_hdr = self.opt_hdr = self.sections = None

    @property
    def closed(self):
        return self.view is None

    ############################################################################

    def __enter__(self):
        return self

    ############################################################################

    def __exit__(self, *args):
        self.close()

    ############################################################################

    def u8(self, offs, idx=0):
        return ctypes.c_uint8.from_address(self.view + offs + idx).value

    ############################################################################

    def u16(self, offs, idx=0):
        return ctypes.c_uint16.from_address(self.view + offs + idx * 2).value

    ############################################################################

    def u32(self, offs, idx=0):
        return ctypes.c_uint32.from_address(self.view + offs + idx * 4).value

    ############################################################################

    def u64(self, offs, idx=0):
        return ctypes.c_uint64.from_address(self.view + offs + idx * 8).value

    ############################################################################

    _std_uint_getters = {1: u8, 2: u16, 4: u32, 8: u64}

    def uint(self, size, offs, idx=0):
        if getter := self._std_uint_getters.get(size, None):
            return getter(self, offs, idx)
        addr = self.view + offs + idx * size
        return int.from_bytes(ctypes.string_at(addr, size), byteorder="little")

    ############################################################################

    def ctypes_obj(self, cls, offs, idx=0):
        return cls.from_address(self.view + offs + idx * ctypes.sizeof(cls))

    ############################################################################

    def __getitem__(self, key):
        if isinstance(key, int):
            if key < 0:
                key += self.size
            if 0 <= key < self.size:
                return ctypes.c_uint8.from_address(self.view + key).value
            raise IndexError
        if isinstance(key, slice):
            start, stop, step = key.indices(self.size)
            return ctypes.string_at(self.view + start, stop - start)[::step]
        raise TypeError

    ############################################################################

    def bstring(self, offs, enc="ascii", err="backslashreplace"):
        return ctypes.string_at(self.view + offs).decode(enc, err)

    ############################################################################

    def string(self, offs):
        return ctypes.wstring_at(self.view + offs)

    ############################################################################

    def _file_header_addr(self):
        if self.u16(self.E_MAGIC) == IMAGE_DOS_SIGNATURE:
            offs = self.u32(self.E_LFANEW)
            if self.u32(offs) == IMAGE_NT_SIGNATURE:
                # len(IMAGE_NT_SIGNATURE) == 4
                return self.view + offs + 4
        return None

    ############################################################################

    def section_from_rva(self, rva):
        for s in self.sections:
            size = s.VirtualSize or s.SizeOfRawData
            if s.VirtualAddress <= rva < s.VirtualAddress + size:
                return s
        return None

    ############################################################################

    def offs_from_rva(self, rva):
        if s := self.section_from_rva(rva):
            return rva - (s.VirtualAddress - s.PointerToRawData)
        return None

    ############################################################################

    def img_dir(self, idx):
        if 0 <= idx < self.opt_hdr.NumberOfRvaAndSizes:
            dd = self.opt_hdr.DataDirectory[idx]
            return dd.VirtualAddress, dd.Size
        raise IndexError

    ############################################################################

    def is_wow(self):
        host_arch = kernel.get_wow64_info(kernel.GetCurrentProcess())[0]
        return host_arch != self.file_hdr.Machine

################################################################################

class API_SET_NAMESPACE(SizedStruct):
    _fields_ = (
        ("Version", ULONG),
        ("Size", ULONG),
        ("Flags", ULONG),
        ("Count", ULONG),
        ("EntryOffset", ULONG),
        ("HashOffset", ULONG),
        ("HashFactor", ULONG),
        )

class API_SET_NAMESPACE_ENTRY(SizedStruct):
    _fields_ = (
        ("Flags", ULONG),
        ("NameOffset", ULONG),
        ("NameLength", ULONG),
        ("HashedLength", ULONG),
        ("ValueOffset", ULONG),
        ("ValueCount", ULONG),
        )

class API_SET_VALUE_ENTRY(SizedStruct):
    _fields_ = (
        ("Flags", ULONG),
        ("NameOffset", ULONG),
        ("NameLength", ULONG),
        ("ValueOffset", ULONG),
        ("ValueLength", ULONG),
        )

################################################################################

class ApiSet():
    def __init__(self):
        offs = 0x68 if PTR_HAS_64_BITS else 0x38
        base = PVOID.from_address(ntdll.RtlGetCurrentPeb() + offs).value
        apiset = API_SET_NAMESPACE.from_address(base)
        if apiset.Version == 6:
            self.base = base
            self.count = apiset.Count
            self.entry_addr = base + apiset.EntryOffset
        else:
            self.base = self.count = self.entry_addr = None

    ############################################################################

    def _get_entry_info(self, idx, for_lookup):
        ENTRY_SIZE = ctypes.sizeof(API_SET_NAMESPACE_ENTRY)
        addr = self.entry_addr + idx * ENTRY_SIZE
        entry = API_SET_NAMESPACE_ENTRY.from_address(addr)
        str_len = (
            entry.HashedLength if for_lookup else entry.NameLength
            ) // WCHAR_SIZE
        str_addr = self.base + entry.NameOffset
        return entry, ctypes.wstring_at(str_addr, str_len)

    ############################################################################

    def _enum_values(self, entry):
        value_addr = self.base + entry.ValueOffset
        for _ in range(entry.ValueCount):
            value = API_SET_VALUE_ENTRY.from_address(value_addr)
            if value.ValueLength:
                res_len = value.ValueLength // WCHAR_SIZE
                res_addr = self.base + value.ValueOffset
                yield ctypes.wstring_at(res_addr, res_len)
            value_addr += ctypes.sizeof(API_SET_VALUE_ENTRY)

    ############################################################################

    def enum_entries(self):
        if self.base:
            for i in range(self.count):
                entry, name = self._get_entry_info(i, False)
                yield (name, list(self._enum_values(entry)))

    ############################################################################

    def lookup(self, dllname):
        if self.base:
            dllname = dllname.lower()
            # entries are sorted -> binary search
            mini = 0
            maxi = self.count - 1
            while mini <= maxi:
                curi = (mini + maxi) // 2
                entry, name = self._get_entry_info(curi, True)
                if dllname.startswith(name):
                    if lst := list(self._enum_values(entry)):
                        return lst[-1]
                    break
                if dllname < name:
                    maxi = curi - 1
                else:
                    mini = curi + 1

        return ""

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# see here for equivalent C++ sample code
# https://github.com/microsoft/Windows-classic-samples/blob/main/Samples/Hyper-V/Storage/cpp/AttachVirtualDisk.cpp

import sys
from ctwin32 import virtdisk as vdsk

path = sys.argv[1]
do_detach = len(sys.argv) > 2

oparams = vdsk.OPEN_VIRTUAL_DISK_PARAMETERS()
oflags = vdsk.OPEN_VIRTUAL_DISK_FLAG_NONE

# Specify UNKNOWN for both device and vendor so the system will use
# the file extension to determine the correct VHD format. The default values
# do exactly that.
storage_type = vdsk.VIRTUAL_STORAGE_TYPE()

if (ext := path.rsplit(".", 1)[1].lower()) == "iso":
    oparams.Version = vdsk.OPEN_VIRTUAL_DISK_VERSION_1
    acc_mask = vdsk.VIRTUAL_DISK_ACCESS_READ
    aflags = (
        vdsk.ATTACH_VIRTUAL_DISK_FLAG_PERMANENT_LIFETIME
        | vdsk.ATTACH_VIRTUAL_DISK_FLAG_READ_ONLY
        )
else:
    oparams.Version = vdsk.OPEN_VIRTUAL_DISK_VERSION_2
    oparams.Version2.GetInfoOnly = False
    aflags = vdsk.ATTACH_VIRTUAL_DISK_FLAG_PERMANENT_LIFETIME

    # not tested -> CHANGE THIS
    acc_mask = vdsk.VIRTUAL_DISK_ACCESS_NONE

with vdsk.OpenVirtualDisk(storage_type, path, acc_mask, oflags, oparams) as vd:
    if do_detach:
        vdsk.DetachVirtualDisk(vd)
    else:
        vdsk.AttachVirtualDisk(vd, aflags)



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import argparse
from ctwin32 import (
    ntdll,
    user,
    powrprof,
    wtsapi,
    EWX_POWEROFF,
    EWX_REBOOT,
    EWX_LOGOFF,
    SE_SHUTDOWN_PRIVILEGE,
    WTSActive
    )

################################################################################

def parse_args():
    ape = argparse.ArgumentParser()
    grp = ape.add_mutually_exclusive_group(required=True)
    grp.add_argument("-s", "--shutdown", action="store_true")
    grp.add_argument("-r", "--restart", action="store_true")
    grp.add_argument("-l", "--logout", action="store_true")
    grp.add_argument("-e", "--hibernate", action="store_true")
    grp.add_argument("-y", "--standby", action="store_true")
    grp.add_argument("-k", "--lock", action="store_true")
    grp.add_argument("-d", "--disconnect", action="store_true")
    return ape.parse_args()

################################################################################

def disconnect_rdp():
    for info in wtsapi.WTSEnumerateSessions():
        if info.State == WTSActive:
            if info.pWinStationName.startswith("RDP-Tcp"):
                wtsapi.WTSDisconnectSession(info.SessionId)

################################################################################

def main():
    args = parse_args()

    if args.shutdown or args.restart:
        try:
            ntdll.RtlAdjustPrivilege(SE_SHUTDOWN_PRIVILEGE, True)
        except OSError as e:
            print(e, file=sys.stderr)
            sys.exit(1)

    if args.shutdown:
        user.ExitWindowsEx(EWX_POWEROFF, 0)

    elif args.restart:
        user.ExitWindowsEx(EWX_REBOOT, 0)

    elif args.logout:
        user.ExitWindowsEx(EWX_LOGOFF, 0)

    elif args.hibernate:
        powrprof.SetSuspendState(True, False, False)

    elif args.standby:
        powrprof.SetSuspendState(False, False, False)

    elif args.lock:
        user.LockWorkStation()

    elif args.disconnect:
        disconnect_rdp()

################################################################################

if __name__ == "__main__":
    main()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This sample was inspired by
# https://www.codeproject.com/Articles/5363023/How-to-Restart-a-USB-Port
#
################################################################################

import sys
import ctypes
from ctwin32 import (
    advapi,
    cfgmgr,
    kernel,
    ntdll,
    CTL_CODE,
    FILE_ANY_ACCESS,
    FILE_DEVICE_UNKNOWN,
    METHOD_BUFFERED,
    SPDRP_LOCATION_INFORMATION,
    )
from ctwin32.wtypes import (
    GUID,
    ULONG,
    )

################################################################################

class USB_CYCLE_PORT_PARAMS(ctypes.Structure):
    _fields_ = (
        ("ConnectionIndex", ULONG),
        ("StatusReturned", ULONG),
        )

################################################################################

def restart_usb_port(device_id):
    # Step 1: find the USB device in the device manager
    devinst = cfgmgr.CM_Locate_DevNode(device_id)

    # Step 2: Determine the USB port number.
    # SPDRP codes are zero based while CM_DRP codes are one based
    prop = SPDRP_LOCATION_INFORMATION + 1
    location, _ = cfgmgr.CM_Get_DevNode_Registry_Property(devinst, prop)
    # must be like "Port_#0004.Hub_#0014"
    try:
        portnumber = int(location.removeprefix("Port_#").split(".")[0])
    except ValueError as e:
        e.args = (f"{device_id} -> not a direct child of a hub",)
        raise

    # Step 3: the USB hub is the parent device
    hub = cfgmgr.CM_Get_Parent(devinst)

    # Step 4: request the USB hub's device interface
    didstr = cfgmgr.CM_Get_Device_ID(hub)
    guid = GUID("f18a0e88-c30c-11d0-8815-00a0c906bed8")
    path = cfgmgr.CM_Get_Device_Interface_List(guid, didstr)[0]

    # Step 5: open the hub
    with kernel.create_file(path) as hhub:

        # Step 6: call IOCTL_USB_HUB_CYCLE_PORT
        FILE_DEVICE_USB = FILE_DEVICE_UNKNOWN
        USB_HUB_CYCLE_PORT = 273
        ioctl = CTL_CODE(
            FILE_DEVICE_USB,
            USB_HUB_CYCLE_PORT,
            METHOD_BUFFERED,
            FILE_ANY_ACCESS
            )
        olen = ctypes.sizeof(USB_CYCLE_PORT_PARAMS)
        params = USB_CYCLE_PORT_PARAMS(portnumber, 0)
        buf = kernel.DeviceIoControl(hhub, ioctl, params, olen)

    result = USB_CYCLE_PORT_PARAMS.from_buffer(buf)
    ntdll.raise_failed_status(result.StatusReturned)

################################################################################

if __name__ == "__main__":
    if advapi.running_as_admin():
        # first arg has to be a `device ID`, e.g.
        # "USB\VID_046D&PID_0825\0FE2C940" (a direct child of a hub)
        # since this contains a `&` it needs to be quoted on the command line.
        restart_usb_port(sys.argv[1])
    else:
        print("restarting an USB port requires administrative privileges.")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    DWORD,
    PCHAR,
    PDWORD,
    POINTER,
    PWSTR,
    ScdToBeClosed,
    UINT,
    ULONG,
    )
from . import (
    ref,
    raise_on_err,
    suppress_winerr,
    fun_fact,
    ERROR_MORE_DATA,
    ERROR_NO_MORE_ITEMS
    )

_msi = ctypes.WinDLL("msi.dll", use_last_error=True)

MSIDBOPEN_READONLY     = PWSTR(0)
MSIDBOPEN_TRANSACT     = PWSTR(1)
MSIDBOPEN_DIRECT       = PWSTR(2)
MSIDBOPEN_CREATE       = PWSTR(3)
MSIDBOPEN_CREATEDIRECT = PWSTR(4)
MSIDBOPEN_PATCHFILE    = PWSTR(32 // 2)

BIN_NAME_IDX = 1  # field index of binary name
BIN_DATA_IDX = 2  # field index of binary data

################################################################################

_MsiCloseHandle = fun_fact(_msi.MsiCloseHandle, (UINT, ULONG))

def MsiCloseHandle(hdl):
    raise_on_err(_MsiCloseHandle(hdl))

################################################################################

class MSIHANDLE(ScdToBeClosed, ULONG, close_func=MsiCloseHandle, invalid=0):
    pass

PMSIHANDLE = POINTER(MSIHANDLE)

################################################################################

_MsiOpenDatabase = fun_fact(
    _msi.MsiOpenDatabaseW, (UINT, PWSTR, PWSTR, PMSIHANDLE)
    )

def MsiOpenDatabase(filename, persist):
    hdl = MSIHANDLE()
    raise_on_err(_MsiOpenDatabase(filename, persist, ref(hdl)))
    return hdl

################################################################################

_MsiDatabaseOpenView = fun_fact(
    _msi.MsiDatabaseOpenViewW, (UINT, MSIHANDLE, PWSTR, PMSIHANDLE)
    )

def MsiDatabaseOpenView(dbase, query):
    hdl = MSIHANDLE()
    raise_on_err(_MsiDatabaseOpenView(dbase, query, ref(hdl)))
    return hdl

################################################################################

_MsiViewExecute = fun_fact(
    _msi.MsiViewExecute, (UINT, MSIHANDLE, MSIHANDLE)
    )

def MsiViewExecute(view, record):
    raise_on_err(_MsiViewExecute(view, record))

################################################################################

_MsiViewFetch = fun_fact(
    _msi.MsiViewFetch, (UINT, MSIHANDLE, PMSIHANDLE)
    )

def MsiViewFetch(view):
    hdl = MSIHANDLE()
    raise_on_err(_MsiViewFetch(view, ref(hdl)))
    return hdl

################################################################################

def view_enum_records(view):
    with suppress_winerr(ERROR_NO_MORE_ITEMS):
        while True:
            with MsiViewFetch(view) as record:
                yield record

################################################################################

_MsiRecordGetFieldCount = fun_fact(
    _msi.MsiRecordGetFieldCount, (UINT, MSIHANDLE)
    )

def MsiRecordGetFieldCount(record):
    return _MsiRecordGetFieldCount(record)

################################################################################

_MsiRecordGetString = fun_fact(
    _msi.MsiRecordGetStringW, (
        UINT,
        MSIHANDLE,
        UINT,
        PWSTR,
        PDWORD
        )
    )

def MsiRecordGetString(record, field_idx):
    size = DWORD(512)
    err = ERROR_MORE_DATA
    while err == ERROR_MORE_DATA:
        buf = string_buffer(size.value)
        err = _MsiRecordGetString(record, field_idx, buf, ref(size))
    raise_on_err(err)
    return buf.value

################################################################################

_MsiRecordReadStream = fun_fact(
    _msi.MsiRecordReadStream, (
        UINT,
        MSIHANDLE,
        UINT,
        PCHAR,
        PDWORD
        )
    )

def MsiRecordReadStream(record, field_idx, size):
    res = byte_buffer(size)
    size = DWORD(size)
    raise_on_err(_MsiRecordReadStream(record, field_idx, res, ref(size)))
    return res.raw[:size.value]

################################################################################

def record_read_stream_all(record, field_idx):
    chunk_size = 64 * 1024
    chunks = []
    chunk = MsiRecordReadStream(record, field_idx, chunk_size)
    while chunk:
        chunks.append(chunk)
        chunk = MsiRecordReadStream(record, field_idx, chunk_size)
    return b"".join(chunks)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from ctwin32 import (
    advapi,
    kernel,
    suppress_winerr,
    ERROR_FILE_NOT_FOUND,
    ERROR_INVALID_FUNCTION,
    ERROR_MORE_DATA,
    ERROR_NOT_SUPPORTED,
    GENERIC_EXECUTE,
    )
from ctwin32.wtypes import ULONG, LARGE_INTEGER

################################################################################

IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS = 0x00560000

class DISK_EXTENT(ctypes.Structure):
    _fields_ = (
        ("DiskNumber", ULONG),
        ("StartingOffset", LARGE_INTEGER),
        ("ExtentLength", LARGE_INTEGER),
        )

def get_disk_extends(buf):

    num = ULONG.from_buffer(buf).value

    class VOLUME_DISK_EXTENTS(ctypes.Structure):
        _fields_ = (
            ("NumberOfDiskExtents", ULONG),
            ("Extents", DISK_EXTENT * num),
            )
    return VOLUME_DISK_EXTENTS.from_buffer(buf).Extents

################################################################################

def get_volume_extends(vol_name):
    with kernel.create_file(vol_name, GENERIC_EXECUTE) as hdl:
        size = 128
        while True:
            size *= 2
            with suppress_winerr(ERROR_MORE_DATA):
                buf = kernel.DeviceIoControl(
                    hdl,
                    IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS,
                    None,
                    size
                    )
                break
    return get_disk_extends(buf)

################################################################################

def get_disks(vol, dev, pth):
    if advapi.running_as_admin():
        # while all volumes have names, that can be opened, this requires
        # administrative privileges
        name = vol[:-1]
    elif pth:
        name = rf"\\.\{pth[0][:-1]}"
    else:
        name = dev.replace(r"\Device", r"\\?")

    try:
        extends = get_volume_extends(name)
    except OSError as e:
        if e.winerror in (ERROR_INVALID_FUNCTION, ERROR_NOT_SUPPORTED):
            return "not supported"
        return e.strerror

    return ", ".join(str(e.DiskNumber) for e in extends)

################################################################################

for volume in kernel.enum_volumes():
    # strip '\\?\' prefix and trailing '\'
    device = kernel.QueryDosDevice(volume[4:-1])
    paths = []
    with suppress_winerr(ERROR_FILE_NOT_FOUND):
        paths = kernel.GetVolumePathNamesForVolumeName(volume)
    disks = get_disks(volume, device, paths)
    paths = ", ".join(paths)

    print(f"\nvolume: {volume}")
    print(f"    device: {device}")
    print(f"    paths: {paths}")
    print(f"    disks: {disks}")



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import argparse
from ctwin32.bcrypt import (
    BCRYPT_ECDSA_P256_ALGORITHM,
    BCRYPT_SHA256_ALGORITHM,
    BCRYPT_ECCPRIVATE_BLOB,
    BCRYPT_ECCPUBLIC_BLOB,
    BCryptExportKey,
    BCryptOpenAlgorithmProvider,
    BCryptHash,
    BCryptGenerateKeyPair,
    BCryptFinalizeKeyPair,
    BCryptImportKeyPair,
    BCryptSignHash,
    BCryptVerifySignature,
    )

################################################################################

SIGN_ALGO      = BCRYPT_ECDSA_P256_ALGORITHM
SIGN_ALGO_BITS = 256
HASH_ALGORITHM = BCRYPT_SHA256_ALGORITHM
BLOB_PRIVATE   = BCRYPT_ECCPRIVATE_BLOB
BLOB_PUBLIC    = BCRYPT_ECCPUBLIC_BLOB

################################################################################

def export_key(key, btype, file_name):
    with open(file_name, "wb") as f:
        f.write(BCryptExportKey(key, btype))

################################################################################

def hash_file(file_name):
    with BCryptHash(HASH_ALGORITHM) as hash:
        chunk = 0x10000
        with open(file_name, "rb") as f:
            while data := f.read(chunk):
                hash.update(data)
            return hash.digest()

################################################################################

def create_key_pair(private_name, public_name):
    with BCryptOpenAlgorithmProvider(SIGN_ALGO) as alg:
        with BCryptGenerateKeyPair(alg, SIGN_ALGO_BITS) as key:
            BCryptFinalizeKeyPair(key)
            export_key(key, BLOB_PRIVATE, private_name)
            export_key(key, BLOB_PUBLIC, public_name)

################################################################################

def sign(source_name, key_file, sig_file, out_hex):
    with BCryptOpenAlgorithmProvider(SIGN_ALGO) as alg:
        with open(key_file, "rb") as f:
            kbuf = f.read()
        with BCryptImportKeyPair(alg, BLOB_PRIVATE, kbuf) as key:
            signature = BCryptSignHash(key, hash_file(source_name))
            if out_hex:
                print(signature.hex())
            else:
                with open(sig_file, "wb") as f:
                    f.write(signature)

################################################################################

def verify(source_name, key_file, sig_file):
    with BCryptOpenAlgorithmProvider(SIGN_ALGO) as alg:
        with open(key_file, "rb") as f:
            kbuf = f.read()
        with BCryptImportKeyPair(alg, BLOB_PUBLIC, kbuf) as key:
            with open(sig_file, "rb") as f:
                signature = f.read()
            ok = BCryptVerifySignature(key, hash_file(source_name), signature)
            print(
                "The signature is valid." if ok
                else "Signature verification failed."
                )

################################################################################

def parse_args():
    ape = argparse.ArgumentParser()
    sub = ape.add_subparsers(
        title="available commands",
        dest="cmd",
        help="action",
        metavar="command"
        )

    create = sub.add_parser("create", help="create a new key pair")
    create.add_argument(
        dest="private",
        metavar="<private_keyfile>",
        help="name of private key file"
        )
    create.add_argument(
        dest="public",
        metavar="<public_keyfile>",
        help="name of public key file"
        )

    sign = sub.add_parser("sign", help="create signature for a file")
    sign.add_argument(
        "-k",
        dest="keyfile",
        metavar="<private_keyfile>",
        help="name of private key file",
        required=True
        )
    sign.add_argument(
        "-s",
        dest="sigfile",
        metavar="<output_sigfile>",
        help="name of output file for signature",
        required=True
        )
    sign.add_argument(
        "-x",
        dest="outhex",
        help="output signature as hex string",
        action="store_true"
        )
    sign.add_argument(
        dest="sourcefile",
        metavar="<source_file>",
        help="name of file to sign"
        )

    verify = sub.add_parser("verify", help="verify signature for a file")
    verify.add_argument(
        "-k",
        dest="keyfile",
        metavar="<public_keyfile>",
        help="name of public key file",
        required=True
        )
    verify.add_argument(
        "-s",
        dest="sigfile",
        metavar="<input_sigfile>",
        help="name of signature file",
        required=True
        )
    verify.add_argument(
        dest="sourcefile",
        metavar="<source_file>",
        help="name of file to verify"
        )

    return ape.parse_args()

################################################################################

def main():
    args = parse_args()
    if args.cmd == "create":
        create_key_pair(args.private, args.public)
    elif args.cmd == "sign":
        sign(args.sourcefile, args.keyfile, args.sigfile, args.outhex)
    elif args.cmd == "verify":
        verify(args.sourcefile, args.keyfile, args.sigfile)

################################################################################

if __name__ == "__main__":
    main()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This sample retrieves and prints power request. This is similar to what
# `powercfg.exe /requests` does.

import ctypes
from ctwin32 import (
    advapi,
    kernel,
    powrprof,
    user,
    suppress_winerr,
    ERROR_INSUFFICIENT_BUFFER,
    LOAD_LIBRARY_AS_DATAFILE,
    LOAD_LIBRARY_AS_IMAGE_RESOURCE,
    )
from ctwin32.wtypes import (
    INT,
    SIZE_T,
    ULONG,
    ULONG_PTR,
    USHORT,
    )

################################################################################

# flags for COUNTED_REASON_CONTEXT_RELATIVE.Flags
DIAGNOSTIC_REASON_SIMPLE_STRING = 0x00000001
DIAGNOSTIC_REASON_DETAILED_STRING = 0x00000002
DIAGNOSTIC_REASON_NOT_SPECIFIED = 0x80000000
DIAGNOSTIC_REASON_INVALID_FLAGS = ~ 0x80000007

class _COUNTED_REASON_CTXT1(ctypes.Structure):
    _fields_ = (
        ("ResourceFileNameOffset", ULONG_PTR),
        ("ResourceReasonId", USHORT),
        ("StringCount", ULONG),
        ("SubstitutionStringsOffset", ULONG),
        )

class _COUNTED_REASON_CTXT2(ctypes.Union):
    _anonymous_ = ("_anon1",)
    _fields_ = (
        ("_anon1", _COUNTED_REASON_CTXT1),
        ("SimpleStringOffset", ULONG_PTR),
        )

class COUNTED_REASON_CONTEXT_RELATIVE(ctypes.Structure):
    _anonymous_ = ("_anon1",)
    _fields_ = (
        ("Flags", ULONG),
        ("_anon1", _COUNTED_REASON_CTXT2),
        )

# caller types for DIAGNOSTIC_BUFFER.CallerType
KERNEL_REQUESTER = 0
PROCESS_REQUESTER = 1
SERVICE_REQUESTER = 2

class _DIAG_BUFF1(ctypes.Structure):
    _fields_ = (
        ("ProcessImageNameOffset", ULONG_PTR),
        ("ProcessId", ULONG),
        ("ServiceTag", ULONG),
        )

class _DIAG_BUFF2(ctypes.Structure):
    _fields_ = (
        ("DeviceDescriptionOffset", ULONG_PTR),
        ("DevicePathOffset", ULONG_PTR),
        )

class _DIAG_BUFF3(ctypes.Union):
    _anonymous_ = ("_anon1", "_anon2")
    _fields_ = (
        ("_anon1", _DIAG_BUFF1),
        ("_anon2", _DIAG_BUFF2),
        )

class DIAGNOSTIC_BUFFER(ctypes.Structure):
    _anonymous_ = ("_anon3",)
    _fields_ = (
        ("Size", SIZE_T),
        ("CallerType", INT),
        ("_anon3", _DIAG_BUFF3),
        ("ReasonOffset", ULONG_PTR),
        )

# !! for simplicity we ignore all the definitions older than Windows 10 RS1+ !!

POWER_REQUEST_SUPPORTED_TYPES = 6

class POWER_REQUEST(ctypes.Structure):
    _fields_ = (
        ("SupportedRequestMask", ULONG),
        ("PowerRequestCount", ULONG * POWER_REQUEST_SUPPORTED_TYPES),
        ("DiagnosticBuffer", DIAGNOSTIC_BUFFER),
        )

################################################################################

def load_res_str(mod_name, msg_id):
    LOLI_FLAGS = LOAD_LIBRARY_AS_DATAFILE | LOAD_LIBRARY_AS_IMAGE_RESOURCE
    with kernel.LoadLibraryEx(mod_name, LOLI_FLAGS) as hmod:
        return user.LoadString(hmod, msg_id)

################################################################################

request_types = {
    0: "DISPLAY",
    1: "SYSTEM",
    2: "AWAYMODE",
    3: "EXECUTION",
    4: "PERFBOOST",
    5: "ACTIVELOCKSCREEN",
    }

requester_types = {
    KERNEL_REQUESTER:  "[DRIVER]",
    PROCESS_REQUESTER: "[PROCESS]",
    SERVICE_REQUESTER: "[SERVICE]",
    }

################################################################################

def get_power_requests():
    GetPowerRequestList = 45
    size = 2048
    buf = None
    while buf is None:
        size *= 2
        with suppress_winerr(ERROR_INSUFFICIENT_BUFFER):
            buf = powrprof.PowerInformationWithPrivileges(
                GetPowerRequestList,
                size
                )

    count = ULONG.from_buffer(buf).value

    class POWER_REQUEST_LIST(ctypes.Structure):
        _fields_ = (
            ("Count", ULONG),
            ("PowerRequestOffsets", ULONG_PTR * count),
            )
    prl = POWER_REQUEST_LIST.from_buffer(buf)

    requests = {k: [] for k in request_types}
    for ri in range(prl.Count):
        pr = POWER_REQUEST.from_buffer(buf, prl.PowerRequestOffsets[ri])

        for rt in request_types:
            # Must not access elements of `PowerRequestCount` that are not
            # approved by `SupportedRequestMask`. Furthermore only requests
            # that have `PowerRequestCount[rt]` != 0 are active.
            add_this = (
                ((1 << rt) & pr.SupportedRequestMask) and
                pr.PowerRequestCount[rt]
                )
            if not add_this:
                continue

            calltype = pr.DiagnosticBuffer.CallerType
            callt_str = requester_types[calltype]
            db = pr.DiagnosticBuffer
            db_addr = ctypes.addressof(db)
            if calltype == KERNEL_REQUESTER:
                callid = [
                    ctypes.wstring_at(db_addr + db.DeviceDescriptionOffset),
                    ctypes.wstring_at(db_addr + db.DevicePathOffset)
                    ]
            else:
                callid = [
                    ctypes.wstring_at(db_addr + db.ProcessImageNameOffset),
                    f"pid:{db.ProcessId}"
                    ]
            reason = ""
            if pr.DiagnosticBuffer.ReasonOffset:
                ro = COUNTED_REASON_CONTEXT_RELATIVE.from_address(
                    db_addr + pr.DiagnosticBuffer.ReasonOffset
                    )
                ro_addr = ctypes.addressof(ro)
                if ro.Flags & DIAGNOSTIC_REASON_SIMPLE_STRING:
                    reason = ctypes.wstring_at(ro_addr + ro.SimpleStringOffset)
                elif ro.Flags & DIAGNOSTIC_REASON_DETAILED_STRING:
                    mod_name = ctypes.wstring_at(
                        ro_addr + ro.ResourceFileNameOffset
                        )
                    reason = load_res_str(mod_name, ro.ResourceReasonId)

            requests[rt].append((callt_str, callid, reason))

    return requests

################################################################################

def print_power_requests():
    requests = get_power_requests()
    print()
    for req_type in sorted(request_types):
        requesters = requests[req_type]
        tstr = request_types[req_type]
        print(f"{tstr}:\n{(len(tstr) + 1) * '-'}")
        if not requesters:
            print("None\n")
            continue
        for ct, cid, reason in requesters:
            print(f"{ct} ", end="")
            for element in cid:
                print(element)
            print(f"{reason}\n")

################################################################################

if __name__ == "__main__":

    if advapi.running_as_admin():
        print("initial:")
        print(80 * "-")
        print_power_requests()

        # create and activate a SystemRequired request
        print("with self created 'SystemRequired' request:")
        print(80 * "-")
        reason = "demonstrating power requests"
        req = kernel.PowerRequest.SystemRequired
        with kernel.create_power_request(reason, req) as hdl:
            print_power_requests()

        print("final:")
        print(80 * "-")
        print_power_requests()
    else:
        print("Inquiring power request requires administrative privileges.")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This script can list imported and exported functions of executable PE files.
#
################################################################################

from argparse import ArgumentParser
from contextlib import suppress
from pathlib import Path

from ctwin32 import (
    IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT,
    IMAGE_DIRECTORY_ENTRY_EXPORT,
    IMAGE_DIRECTORY_ENTRY_IMPORT,
    IMAGE_ORDINAL_FLAG32,
    IMAGE_ORDINAL_FLAG64,
    )
from ctwin32.pemap import (
    ApiSet,
    dlattrRva,
    IMAGE_EXPORT_DIRECTORY,
    IMAGE_IMPORT_DESCRIPTOR,
    IMAGE_DELAYLOAD_DESCRIPTOR,
    NotPeError,
    pemap,
    )

################################################################################

API_SET = ApiSet()

PM_STANDARD = 0
PM_EXP_ONLY = 1
PM_IMP_ONLY = 2
PM_IMP_ONLY_MOD_ONLY = 3
PM_IMP_OF_MOD_ONLY = 4

################################################################################

def name_in_names(name, names, src, indent, extra=None):
    for n in names:
        if n == "*" or n.lower() == name.lower():
            iname = f"{indent * ' '}{name}"
            return f"{iname} ({src} {extra})" if extra else f"{iname} ({src})"
    return None

################################################################################

def find_imp_funcs_one_mod(pe, offs_INT, names, src):
    result = []
    idx = 0
    if pe.is64bit:
        get_rva = pe.u64
        mask = IMAGE_ORDINAL_FLAG64
    else:
        get_rva = pe.u32
        mask = IMAGE_ORDINAL_FLAG32
    while True:
        rva = get_rva(offs_INT, idx)
        idx += 1
        if not rva:
            break
        if (mask & rva) == 0:
            # IMAGE_IMPORT_BY_NAME.Name
            name = pe.bstring(pe.offs_from_rva(rva) + 2)
        else:
            name = f"#{rva & 0xffff}"
        if found := name_in_names(name, names, src, 8):
            result.append(found)

    return result

################################################################################

def find_imports_one_mod(pe, mode, names, rva_name, rva_INT, src):
    mod_name = pe.bstring(pe.offs_from_rva(rva_name))
    mod_eq_n0 = names and mod_name.lower() == names[0].lower()
    if lu := API_SET.lookup(mod_name):
        mod_name = f"{mod_name} -> {lu}"
    if mode == PM_IMP_OF_MOD_ONLY:
        if mod_eq_n0:
            names = ["*"]
        else:
            return []
    elif mode == PM_IMP_ONLY_MOD_ONLY:
        return [f"    {src} {mod_name}"]

    if offs_INT := pe.offs_from_rva(rva_INT):
        if mod_res := find_imp_funcs_one_mod(pe, offs_INT, names, src):
            return [f"    {mod_name}", *mod_res]
    return []

################################################################################

def find_imports(pe, mode, names):
    result = []
    src = "import"
    im_rva, _ = pe.img_dir(IMAGE_DIRECTORY_ENTRY_IMPORT)
    if not im_rva:
        return result
    iid_offs = pe.offs_from_rva(im_rva)
    idx = 0
    while True:
        iid = pe.ctypes_obj(IMAGE_IMPORT_DESCRIPTOR, iid_offs, idx)
        idx += 1
        if iid.TimeDateStamp == 0 and iid.Name == 0:
            # end of imports
            break

        result.extend(
            find_imports_one_mod(
                pe,
                mode,
                names,
                iid.Name,
                iid.OriginalFirstThunk or iid.FirstThunk,
                src
                )
            )

    return result

################################################################################

def find_delay_imports(pe, mode, names):
    result = []
    src = "delay import"
    rva, _ = pe.img_dir(IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT)
    if not rva:
        return result
    offs = pe.offs_from_rva(rva)
    if not offs:
        return result
    idx = 0
    while True:
        idd = pe.ctypes_obj(IMAGE_DELAYLOAD_DESCRIPTOR, offs, idx)
        idx += 1
        if idd.DllNameRVA == 0 or (idd.Attributes & dlattrRva) == 0:
            break
        result.extend(
            find_imports_one_mod(
                pe,
                mode,
                names,
                idd.DllNameRVA,
                idd.ImportNameTableRVA,
                src
                )
            )

    return result

################################################################################

def find_exports(pe, names):
    result = []
    ex_rva, ex_size = pe.img_dir(IMAGE_DIRECTORY_ENTRY_EXPORT)
    ex_end = ex_rva + ex_size
    if not ex_rva:
        return result
    exd = pe.ctypes_obj(IMAGE_EXPORT_DIRECTORY, pe.offs_from_rva(ex_rva))
    func = pe.offs_from_rva(exd.AddressOfFunctions)
    ords = pe.offs_from_rva(exd.AddressOfNameOrdinals)
    fnames = pe.offs_from_rva(exd.AddressOfNames)

    for i in range(exd.NumberOfNames):
        ord_mb = pe.u16(ords, i)
        name = pe.bstring(pe.offs_from_rva(pe.u32(fnames , i)))
        ep = pe.u32(func,  ord_mb)
        fwd = None
        if ex_rva <= ep < ex_end:
            dst = pe.bstring(pe.offs_from_rva(ep))
            fwd = f"forwarder -> {dst}"
        if found := name_in_names(name, names, "export", 4, fwd):
            result.append(found)

    return result

################################################################################

def fimex_in_file(path, mode, names):
    mode_exp = mode in (PM_STANDARD, PM_EXP_ONLY)
    mode_imp = mode in (
        PM_STANDARD,
        PM_IMP_ONLY,
        PM_IMP_ONLY_MOD_ONLY,
        PM_IMP_OF_MOD_ONLY
        )
    with suppress(NotPeError), pemap(path, convert_open_err=True) as pe:
        result = []
        if mode_exp:
            result.extend(find_exports(pe, names))
        if mode_imp:
            ires = find_imports(pe, mode, names)
            ires.extend(find_delay_imports(pe, mode, names))
            if ires and result:
                result.append("")
            result.extend(ires)
        if result:
            result.insert(0, str(path))
            result.append("")
            print("\n".join(result))

################################################################################

def parse_fimex_args():

    ape = ArgumentParser(description="find imports/exports")
    group = ape.add_mutually_exclusive_group()
    g = group.add_argument
    g("-e", "--exp", action="store_true", help="exports only")
    g("-i", "--imp", action="store_true", help="imports only")
    g("-I", "--imo", action="store_true", help="imported modules only")
    g("-m", "--mod", action="store_true", help="imports of module only")
    ape.add_argument("-p", "--path", help="path to search (default: current)")
    help="function names (default: any), in case of -m: module name"
    ape.add_argument("names", nargs="*", help=help)
    args = ape.parse_args()

    mode = PM_STANDARD
    if args.exp:
        mode = PM_EXP_ONLY
    elif args.imp:
        mode = PM_IMP_ONLY
    elif args.imo:
        mode = PM_IMP_ONLY_MOD_ONLY
    elif args.mod:
        mode = PM_IMP_OF_MOD_ONLY
    return Path("." if not args.path else args.path), mode, args.names

################################################################################

def fimex_main():
    path, mode, names = parse_fimex_args()
    if not names:
        names = ["*"]
    for root, _, files in path.walk():
        for f in files:
            fimex_in_file(root / f, mode, names)

################################################################################

if __name__ == "__main__":
    fimex_main()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import time
import ctypes
import msvcrt
from ctwin32 import (
    ntdll,
    ref,
    SystemBasicInformation,
    SystemProcessorPerformanceInformation,
    )

################################################################################

if __name__ == "__main__":
    print(
        "Interrupts per second for the "
        "individual cores and the sum of these:"
        )

    sbi = ntdll.SYSTEM_BASIC_INFORMATION()
    ntdll.raise_failed_status(
        ntdll.NtQuerySystemInformation(
            SystemBasicInformation,
            ref(sbi),
            ctypes.sizeof(sbi),
            None
            )
        )
    num_proc = sbi.NumberOfProcessors

    # last entries are for sums
    last_cnt = (num_proc + 1) * [0]
    cur_cnt = (num_proc + 1) * [0]
    sppi = (ntdll.SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION * num_proc)()

    done_first = False
    while not msvcrt.kbhit():
        ntdll.raise_failed_status(
            ntdll.NtQuerySystemInformation(
                SystemProcessorPerformanceInformation,
                ref(sppi),
                ctypes.sizeof(sppi),
                None
                )
            )
        cur_cnt[num_proc] = 0
        for i in range(num_proc):
            cur_cnt[i] = sppi[i].InterruptCount
            cur_cnt[num_proc] += cur_cnt[i]

        if done_first:
            for i in range(num_proc + 1):
                print(f"{cur_cnt[i] - last_cnt[i]:5} ", end="")
            print("                \r", end="")

        for i in range(num_proc + 1):
            last_cnt[i] = cur_cnt[i]
        done_first = True
        time.sleep(1)

    while msvcrt.kbhit():
        msvcrt.getch()

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# demonstrate device enumerators from setupapi

from ctwin32.setupapi import (
    enum_info_set,
    SetupDiGetClassDevs,
    SetupDiGetDeviceInstanceId,
    get_device_enumerators,
    get_device_classes,
    SetupDiClassNameFromGuid,
    desc_from_info_set,
    )

for e in get_device_enumerators():
    print(f"\nEnumerator {e}:")
    for iset, ddat in enum_info_set(SetupDiGetClassDevs(enumerator=e)):
        try:
            desc = desc_from_info_set(iset, ddat)
        except OSError:
            desc = "UNKNOWN"
        iid = SetupDiGetDeviceInstanceId(iset, ddat)
        print(f"  {desc:50s} {iid}")

for g in get_device_classes():
    cn = SetupDiClassNameFromGuid(g)
    print(f"\nDevice class {g}, {cn}:")
    for iset, ddat in enum_info_set(SetupDiGetClassDevs(guid=g)):
        try:
            desc = desc_from_info_set(iset, ddat)
        except OSError:
            desc = "UNKNOWN"
        iid = SetupDiGetDeviceInstanceId(iset, ddat)
        print(f"  {desc:50s} {iid}")
